

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

int main(void)
{
  int i, j, k;
  int N1, N2, N3;
  double t, T, dt, dx, dy, dz;
  double *u, *u_prev, *tmp;
  double kappa, alpha;
  clock_t start, diff;

  T = N1 = N2 = N3 = 40;
  dx = dy = dz = 1./N1;
  kappa = 0.04;
  dt = (dx*dx)/kappa/6.0;

  u = (double*) malloc(N1*N2*N3*sizeof(double));
  u_prev = (double*) malloc(N1*N2*N3*sizeof(double));
  
  alpha = dt*kappa/(dx*dx);           // = dt*kappa/dy = dt*kappa/dx

  for (i = 0; i <N1*N2*N3; i++) {
    u_prev[i] = i*rand()/1E8;    // some values
  }

  // idea: du/dt = kappa*(d2u/dx2 + ...)
  // discrete: (u_(i, j, k)(t+1) - u_(i, j, k)(t))/dt = 
  //     kappa*((u_(i-1, j, k)(t) - 2*u_(i, j, k) + u(i+1, j, k)(t))/dx*dx + ...)
  // => u_(i, j, k)(t+1) = dt*kappa*(...) + u_(i, j, k)(t)
  
  start = clock();
  t = 0;
  while (t < T) {
    for(i = 1; i < N1-1; i++) {
      for(j = 1; j < N2-1; j++) {
	for(k = 1; k < N3-1; k++) {
	  u[i*N2*N3 + j*N3 + k] = u_prev[i*N2*N3 + j*N3 + k]	\
	    + alpha*(u_prev[(i-1)*N2*N3 + j*N3 + k] +		\
		     u_prev[(i+1)*N2*N3 + j*N3 + k] +		\
		     u_prev[i*N2*N3 + (j-1)*N3 + k] +		\
		     u_prev[i*N2*N3 + (j+1)*N3 + k] +		\
		     u_prev[i*N2*N3 + j*N3 + k-1] +		\
		     u_prev[i*N2*N3 + j*N3 + k+1] -		\
		     6*u_prev[i*N2*N3 + j*N3 + k]);		\
	}
      }
    }
    // print a row (up to 10 numbers) to see difference
    // note that boundary values are constant
    /* for(i = 0; i < N1 && i < 10; i++)
      printf("%f ", u[2*N2*N3 + 2*N2 + i]);
    printf("\n");*/
    
    // double pointer (switch)
    tmp = u_prev;
    u_prev = u;
    u = tmp;

    t += dt;
  }
  diff = clock() - start;
  free(u);
  free(u_prev);
  
  /* FLOPS:
     no. of iterations: T*(N1-2)*(N2-2)*(N3-2)
     pr. iteration: 9 operations (+ - *) */

  long int no_calc = (T/dt)*(N1-2)*(N2-2)*(N3-2)*9;
  printf("FLOPS: %lu\n", no_calc/diff*CLOCKS_PER_SEC);

  return 0;
}
